DICT = '__dict__'
SLOTS = '__slots__'


# def select_values(obj, *keys):
#     return [getattr(obj, k, None) for k in keys]


def select_values(ob, *keys):
    if hasattr(ob, DICT) or hasattr(ob, SLOTS): return [getattr(ob, k, None) for k in keys]
    if isinstance(ob, dict): return [ob[k] if k in ob else None for k in keys]
    if isinstance(ob, list) and (hi := len(ob)) and (lo := -hi):
        return [ob[k] if isinstance(k, int) and lo <= k < hi else None for k in keys]
    return []


def select_entries(ob, keys):
    if hasattr(ob, DICT) or hasattr(ob, SLOTS): return [(k, getattr(ob, k, None)) for k in keys]
    if isinstance(ob, dict): return [(k, ob[k]) if k in ob else None for k in keys]
    if isinstance(ob, list) and (hi := len(ob)) and (lo := -hi):
        return [(k, ob[k]) if isinstance(k, int) and lo <= k < hi else None for k in keys]
    return []


def select_to_dict(ob, *keys):
    if hasattr(ob, DICT) or hasattr(ob, SLOTS): return {k: getattr(ob, k, None) for k in keys}
    if isinstance(ob, dict): return {k: ob[k] if k in ob else None for k in keys}
    if isinstance(ob, list) and (hi := len(ob)) and (lo := -hi):
        return {k: ob[k] if isinstance(k, int) and lo <= k < hi else None for k in keys}
    return {}



DICT = '__dict__'
SLOTS = '__slots__'


def keys(ob):
    return list(ob.__dict__.keys()) if hasattr(ob, DICT) \
        else ob.__slots__ if hasattr(ob, SLOTS) \
        else list(ob.keys()) if isinstance(ob, dict) \
        else [i for i, _ in enumerate(ob)] if isinstance(ob, list) \
        else []


def values(ob):
    return list(ob.__dict__.values()) if hasattr(ob, DICT) \
        else [getattr(ob, s) for s in ob.__slots__] if hasattr(ob, SLOTS) \
        else list(ob.values()) if isinstance(ob, dict) \
        else ob[:] if isinstance(ob, list) \
        else []


def entries(ob):
    return [(s, getattr(ob, s)) for s in ob.__slots__] if hasattr(ob, SLOTS) \
        else list(ob.__dict__.items()) if hasattr(ob, DICT) \
        else list(ob.items()) if isinstance(ob, dict) \
        else list(enumerate(ob)) if isinstance(ob, list) \
        else []


def to_dict(ob):
    return ob.__dict__ if hasattr(ob, DICT) \
        else {s: getattr(ob, s) for s in ob.__slots__} if hasattr(ob, SLOTS) \
        else ob if isinstance(ob, dict) \
        else dict(enumerate(ob)) if isinstance(ob, list) \
        else {}



def iterate_values(entries, vfn):
    for (_, v) in entries: vfn(v)
    pass


def mapper_values(entries, vfn):
    return [(k, vfn(v, i)) for i, (k, v) in enumerate(entries)]


def mutate_values(entries, vfn):
    i, hi = 0, len(entries)
    while i < hi:
        k, v = entries[i]
        entries[i] = (k, vfn(v))
        i += 1
    return entries



import json


def iso(vec: list, val):
    lex = {k: val for k in vec}
    return json.dumps(lex)


def dict_to_json(lex: dict):
    return json.dumps(lex)


def json_to_dict(jso: json):
    return json.loads(jso)



def mutazip(a, b, kfn, vfn=None):
    if vfn is None: vfn = kfn
    hi = len(a)
    while (hi := hi - 1) >= 0:
        ak, av = a[hi]
        bk, bv = b[hi]
        a[hi] = (kfn(ak, bk), vfn(av, bv))
    return a



from typing import Callable, List

from veho.matrix.utils import size


def mutate(mx: List[list], fn: Callable):
    h, w = size(mx)
    x = 0
    while x < h:
        y = 0
        row = mx[x]
        while y < w:
            row[y] = fn(row[y])
            y += 1
        x += 1
    return mx



from veho.vector.zipper.zipper import zipper


def duozipper(a, b, fn):
    return zipper(a, b, fn)


def trizipper(a, b, c, fn):
    return [fn(x, y, z) for (x, y, z) in zip(a, b, c, )]


def quazipper(a, b, c, d, fn):
    return [fn(x1, x2, x3, x4) for (x1, x2, x3, x4) in zip(a, b, c, d)]


def to_duozipper(fn):
    return lambda a, b: duozipper(a, b, fn)


def to_trizipper(fn):
    return lambda a, b, c: trizipper(a, b, c, fn)


def to_quazipper(fn):
    return lambda a, b, c, d: quazipper(a, b, c, d, fn)



from .init import init, iso
from .mapper import iterate, mapper, mutate, shallow
from .margin import margin_mapper, margin_mutate, margin_shallow
from .select import col, row
from .utils import size, transpose



from .mapper import iterate, mapper, mutate
from .margin import margin_mapper, margin_mutate, margin_shallow
from .zipper import duozipper, mutazip, quazipper, to_duozipper, to_quazipper, to_trizipper, trizipper, zipper



def iterate(entries, kfn, vfn=None):
    if vfn is None: vfn = kfn
    for (k, v) in entries: kfn(k), vfn(v)
    pass


def mapper(entries, kfn, vfn=None):
    if vfn is None: vfn = kfn
    return [(kfn(k), vfn(v)) for (k, v) in entries]


def mutate(entries, kfn, vfn=None):
    if vfn is None: vfn = kfn
    i, hi = 0, len(entries)
    while i < hi:
        k, v = entries[i]
        entries[i] = (kfn(k), vfn(v))
        i += 1
    return entries



from veho.columns.getter import column
from veho.matrix import size


def iterate(mx, fn_on_column):
    _, w = size(mx)
    for j in range(w):
        col = column(mx, j)
        fn_on_column(col, j)
    pass


def mapper(mx, fn_on_column):
    _, w = size(mx)
    return [fn_on_column(column(mx, j), j) for j in range(w)]



from veho.matrix.zipper.zipper import zipper


def duozipper(ma, mb, fn):
    return zipper(ma, mb, fn)


def trizipper(ma, mb, mc, fn):
    zipped = list(zip(ma, mb, mc))
    for (i, (ra, rb, rc)) in enumerate(zipped):
        zipped[i] = (fn(a, b, c) for (a, b, c) in zip(ra, rb, rc))
    return zipped


def quazipper(ma, mb, mc, md, fn):
    zipped = list(zip(ma, mb, mc, md))
    for (i, (ra, rb, rc, rd)) in enumerate(zipped):
        zipped[i] = (fn(a, b, c, d) for (a, b, c, d) in zip(ra, rb, rc, rd))
    return zipped


def to_duozipper(fn):
    return lambda ma, mb: duozipper(ma, mb, fn)


def to_trizipper(fn):
    return lambda ma, mb, mc: trizipper(ma, mb, mc, fn)


def to_quazipper(fn):
    return lambda ma, mb, mc, md: quazipper(ma, mb, mc, md, fn)



from veho.entries.enumerate.zipper.zipper import zipper


def duozipper(a, b, kfn, vfn=None):
    return zipper(a, b, kfn, vfn)


def trizipper(a, b, c, kfn, vfn=None):
    if vfn is None: vfn = kfn
    return [(kfn(ak, bk, ck, i), vfn(av, bv, cv, i))
            for i, ((ak, av), (bk, bv), (ck, cv)) in enumerate(zip(a, b, c))]


def quazipper(a, b, c, d, kfn, vfn=None):
    if vfn is None: vfn = kfn
    return [(kfn(ak, bk, ck, dk, i), vfn(av, bv, cv, dv, i))
            for i, ((ak, av), (bk, bv), (ck, cv), (dk, dv)) in enumerate(zip(a, b, c, d))]


def to_duozipper(kfn, vfn=None):
    return lambda a, b: duozipper(a, b, kfn, vfn)


def to_trizipper(kfn, vfn=None):
    return lambda a, b, c: trizipper(a, b, c, kfn, vfn)


def to_quazipper(kfn, vfn=None):
    return lambda a, b, c, d: quazipper(a, b, c, d, kfn, vfn)



def iterate(vec, fn):
    i, hi = 0, len(vec)
    while i < hi:
        fn(vec[i])
        i += 1
    pass

# def iterate(vec, fn):
#     i, hi = 0, len(vec)
#     if (n := length(fn)) == 1:
#         while i < hi:
#             fn(vec[i])
#             i += 1
#     elif n == 2:
#         while i < hi:
#             fn(vec[i], i)
#             i += 1
#     pass



def wind(keys: list, values: list):
    return list(zip(keys, values))


def wind_to_dict(keys: list, values: list):
    return {k: v for k, v in zip(keys, values)}


def iso(keys: list, value):
    return [(key, value) for key in keys]



from .mapper import \
    iterate, iterate_keys, iterate_values, \
    mapper, mapper_keys, mapper_values, \
    mutate, mutate_keys, mutate_values
from .margin import margin_mapper, margin_mutate, margin_shallow
from .zipper import duozipper, mutazip, quazipper, to_duozipper, to_quazipper, to_trizipper, trizipper, zipper
